#pylint: skip-file
# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'data/ui/manager/ui_prune_widget.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_PruneWidget(object):
    def setupUi(self, PruneWidget):
        PruneWidget.setObjectName("PruneWidget")
        PruneWidget.resize(430, 246)
        self.gridLayout_2 = QtWidgets.QGridLayout(PruneWidget)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.prune_monthly_enable = QtWidgets.QCheckBox(PruneWidget)
        self.prune_monthly_enable.setObjectName("prune_monthly_enable")
        self.gridLayout.addWidget(self.prune_monthly_enable, 3, 0, 1, 1)
        self.prune_yearly_enable = QtWidgets.QCheckBox(PruneWidget)
        self.prune_yearly_enable.setObjectName("prune_yearly_enable")
        self.gridLayout.addWidget(self.prune_yearly_enable, 4, 0, 1, 1)
        self.prune_hourly_enable = QtWidgets.QCheckBox(PruneWidget)
        self.prune_hourly_enable.setObjectName("prune_hourly_enable")
        self.gridLayout.addWidget(self.prune_hourly_enable, 0, 0, 1, 1)
        self.prune_daily_enable = QtWidgets.QCheckBox(PruneWidget)
        self.prune_daily_enable.setObjectName("prune_daily_enable")
        self.gridLayout.addWidget(self.prune_daily_enable, 1, 0, 1, 1)
        self.prune_weekly_narchives = QtWidgets.QSpinBox(PruneWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.prune_weekly_narchives.sizePolicy().hasHeightForWidth())
        self.prune_weekly_narchives.setSizePolicy(sizePolicy)
        self.prune_weekly_narchives.setObjectName("prune_weekly_narchives")
        self.gridLayout.addWidget(self.prune_weekly_narchives, 2, 1, 1, 1)
        self.prune_monthly_narchives = QtWidgets.QSpinBox(PruneWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.prune_monthly_narchives.sizePolicy().hasHeightForWidth())
        self.prune_monthly_narchives.setSizePolicy(sizePolicy)
        self.prune_monthly_narchives.setObjectName("prune_monthly_narchives")
        self.gridLayout.addWidget(self.prune_monthly_narchives, 3, 1, 1, 1)
        self.prune_weekly_enable = QtWidgets.QCheckBox(PruneWidget)
        self.prune_weekly_enable.setObjectName("prune_weekly_enable")
        self.gridLayout.addWidget(self.prune_weekly_enable, 2, 0, 1, 1)
        self.prune_yearly_narchives = QtWidgets.QSpinBox(PruneWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.prune_yearly_narchives.sizePolicy().hasHeightForWidth())
        self.prune_yearly_narchives.setSizePolicy(sizePolicy)
        self.prune_yearly_narchives.setObjectName("prune_yearly_narchives")
        self.gridLayout.addWidget(self.prune_yearly_narchives, 4, 1, 1, 1)
        self.prune_daily_narchives = QtWidgets.QSpinBox(PruneWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.prune_daily_narchives.sizePolicy().hasHeightForWidth())
        self.prune_daily_narchives.setSizePolicy(sizePolicy)
        self.prune_daily_narchives.setObjectName("prune_daily_narchives")
        self.gridLayout.addWidget(self.prune_daily_narchives, 1, 1, 1, 1)
        self.prune_hourly_narchives = QtWidgets.QSpinBox(PruneWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.prune_hourly_narchives.sizePolicy().hasHeightForWidth())
        self.prune_hourly_narchives.setSizePolicy(sizePolicy)
        self.prune_hourly_narchives.setObjectName("prune_hourly_narchives")
        self.gridLayout.addWidget(self.prune_hourly_narchives, 0, 1, 1, 1)
        self.gridLayout_2.addLayout(self.gridLayout, 1, 0, 1, 3)
        self.prune_keep_within_enable = QtWidgets.QCheckBox(PruneWidget)
        self.prune_keep_within_enable.setObjectName("prune_keep_within_enable")
        self.gridLayout_2.addWidget(self.prune_keep_within_enable, 2, 0, 1, 1)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.prune_keep_within_amount = QtWidgets.QSpinBox(PruneWidget)
        self.prune_keep_within_amount.setMinimum(1)
        self.prune_keep_within_amount.setMaximum(10000)
        self.prune_keep_within_amount.setObjectName("prune_keep_within_amount")
        self.horizontalLayout_3.addWidget(self.prune_keep_within_amount)
        self.prune_keep_within_unit = QtWidgets.QComboBox(PruneWidget)
        self.prune_keep_within_unit.setObjectName("prune_keep_within_unit")
        self.prune_keep_within_unit.addItem("")
        self.prune_keep_within_unit.addItem("")
        self.prune_keep_within_unit.addItem("")
        self.prune_keep_within_unit.addItem("")
        self.prune_keep_within_unit.addItem("")
        self.prune_keep_within_unit.addItem("")
        self.horizontalLayout_3.addWidget(self.prune_keep_within_unit)
        self.gridLayout_2.addLayout(self.horizontalLayout_3, 2, 1, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(17, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_2.addItem(spacerItem, 2, 2, 1, 1)
        self.prune_enable = QtWidgets.QCheckBox(PruneWidget)
        self.prune_enable.setObjectName("prune_enable")
        self.gridLayout_2.addWidget(self.prune_enable, 0, 0, 1, 2)

        self.retranslateUi(PruneWidget)
        self.prune_keep_within_unit.setCurrentIndex(2)
        QtCore.QMetaObject.connectSlotsByName(PruneWidget)

    def retranslateUi(self, PruneWidget):
        _translate = QtCore.QCoreApplication.translate
        PruneWidget.setWindowTitle(_translate("PruneWidget", "Form"))
        self.prune_monthly_enable.setText(_translate("PruneWidget", "Monthly"))
        self.prune_yearly_enable.setText(_translate("PruneWidget", "Yearly"))
        self.prune_hourly_enable.setText(_translate("PruneWidget", "Hourly"))
        self.prune_daily_enable.setText(_translate("PruneWidget", "Daily"))
        self.prune_weekly_enable.setText(_translate("PruneWidget", "Weekly"))
        self.prune_keep_within_enable.setText(_translate("PruneWidget", "Unconditionally keep archives younger than"))
        self.prune_keep_within_unit.setItemText(0, _translate("PruneWidget", "Minutes"))
        self.prune_keep_within_unit.setItemText(1, _translate("PruneWidget", "Hours"))
        self.prune_keep_within_unit.setItemText(2, _translate("PruneWidget", "Days"))
        self.prune_keep_within_unit.setItemText(3, _translate("PruneWidget", "Weeks"))
        self.prune_keep_within_unit.setItemText(4, _translate("PruneWidget", "Months"))
        self.prune_keep_within_unit.setItemText(5, _translate("PruneWidget", "Years"))
        self.prune_enable.setText(_translate("PruneWidget", "Enable pruning"))




########################################################################
# File name: core.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import ast
import asyncio
import contextlib
import functools
import logging
import pathlib
import signal
import socket
import tempfile
import uuid

from datetime import datetime

import toml

import pytz

import xdg.BaseDirectory

import cuteborg.backend
import cuteborg.subprocess_backend
import cuteborg.config as config

from . import utils, protocol, wctsleep, jobs, devices

if not hasattr(asyncio, "ensure_future"):
    asyncio.ensure_future = asyncio.async


def _schedule_in_interval(now, interval_start, interval_end):
    # TODO: honour scheduling preferences
    if now > interval_start:
        return now
    return interval_start


def _schedule(now, interval_step, interval_unit, last_run):
    start_of_current_interval = utils.align_to_interval(
        now,
        interval_step,
        interval_unit,
    )

    if (last_run is not None and
            last_run >= start_of_current_interval):
        start_of_current_interval = utils.step_interval(
            start_of_current_interval,
            interval_step,
            interval_unit,
        )

    end_of_current_interval = utils.step_interval(
        start_of_current_interval,
        interval_step,
        interval_unit,
    )

    return _schedule_in_interval(
        now,
        start_of_current_interval,
        end_of_current_interval,
    )


@contextlib.contextmanager
def cancelling(fut):
    try:
        yield fut
    finally:
        if not fut.done():
            fut.cancel()


@contextlib.contextmanager
def note_job_block(mapping, key, blocker, notifier):
    unset = object()
    prev = mapping.get(key, unset)
    mapping[key] = blocker
    notifier()
    try:
        yield
    finally:
        if prev is unset:
            try:
                del mapping[key]
            except KeyError:
                pass
        else:
            mapping[key] = prev
        notifier()


class ServerProtocolEndpoint:
    def __init__(self, logger, scheduler):
        super().__init__()
        self.logger = logger
        self.scheduler = scheduler
        self.poll_pending = False
        self.poll_futures = []

    def _delayed_notify_poll(self):
        self.poll_pending = False
        self.logger.debug("notifying pollers")

        for fut in self.poll_futures:
            if fut.done():
                continue

            fut.set_result(None)
        self.poll_futures.clear()

    def notify_poll(self):
        if self.poll_pending:
            return

        self.poll_pending = True
        asyncio.get_event_loop().call_later(
            0.1,
            self._delayed_notify_poll,
        )

    def _handle_status_request(self):
        result = {
            "jobs": [],
        }

        for raw in self.scheduler.get_job_status():
            ((type_, *info),
             job_obj,
             task_fut,
             blocker,
             progress,
             error,
             last_run) = raw

            item = {
                "type": type_,
                "last_run": last_run,
            }

            if type_ == "create_archive":
                job_name, repo_id = info
                item["args"] = {}
                item["args"]["job"] = job_name
                item["args"]["repo"] = {
                    "id": repo_id,
                    "name": self.scheduler.config.repositories[repo_id].name
                }

            elif type_ == "prune":
                repo_id, = info
                item["args"] = {}
                item["args"]["repo"] = {
                    "id": repo_id,
                    "name": self.scheduler.config.repositories[repo_id].name
                }

            item["raw_args"] = list(info)

            if blocker is not None:
                reason, args = blocker
                item["blocking"] = {}
                item["blocking"]["reason"] = reason

                if reason == "waiting_for_repository_lock":
                    repo_id, = args
                    item["blocking"]["struct_info"] = {}
                    item["blocking"]["struct_info"]["repo"] = {
                        "id": repo_id,
                        "name": self.scheduler.config.repositories[repo_id].name
                    }

                elif reason == "sleep_until":
                    dt, = args
                    item["blocking"]["struct_info"] = {}
                    item["blocking"]["struct_info"]["wakeup_at"] = dt

                elif reason == "removable_device":
                    dev_uuid, = args
                    item["blocking"]["struct_info"] = {}
                    item["blocking"]["struct_info"]["device_uuid"] = dev_uuid

                elif reason == "network_host":
                    host, extra = args
                    item["blocking"]["struct_info"] = {}
                    item["blocking"]["struct_info"]["host"] = host
                    if extra is not None:
                        item["blocking"]["struct_info"]["extra"] = extra

                else:
                    item["blocking"]["raw_info"] = list(args)

            if progress is not None:
                item["progress"] = progress

            if error is not None:
                msg, timestamp = error
                item["error"] = {}
                item["error"]["message"] = msg
                item["error"]["since"] = timestamp

            item["running"] = task_fut is not None and not task_fut.done()

            result["jobs"].append(item)

        return protocol.ToplevelCommand.EXTENDED_REQUEST, result

    def _send_status(self, protocol, _):
        try:
            protocol.send(*self._handle_status_request())
        except ConnectionError:
            pass

    def handle_request(self, proto, cmd, extra_data):
        if cmd == protocol.ToplevelCommand.STATUS:
            return self._handle_status_request()
        elif cmd == protocol.ToplevelCommand.RESCHEDULE:
            self.scheduler._trigger_reload_and_reschedule()
            return protocol.ToplevelCommand.OKAY, None
        elif cmd == protocol.ToplevelCommand.POLL:
            self.logger.debug("received poll request")
            fut = asyncio.Future()
            fut.add_done_callback(functools.partial(self._send_status, proto))
            self.poll_futures.append(fut)
            return self._handle_status_request()
        elif cmd == protocol.ToplevelCommand.EXIT:
            self.scheduler._trigger_exit()
            return protocol.ToplevelCommand.OKAY, None
        elif cmd == protocol.ToplevelCommand.EXTENDED_REQUEST:
            try:
                if extra_data["command"] == "force-archive":
                    job = extra_data["force-archive"]["name"]
                    only = extra_data["force-archive"].get("repositories")
                    self.scheduler._force_archive(
                        job,
                        only=only,
                    )
                    return protocol.ToplevelCommand.OKAY, None
                elif extra_data["command"] == "force-prune":
                    self.scheduler._force_prune(
                        extra_data["force-prune"]["repositories"]
                    )
                    return protocol.ToplevelCommand.OKAY, None
            except KeyError as exc:
                self.logger.debug(
                    "missing required data for extended request (%r)",
                    exc,
                )

        self.logger.debug(
            "no way to handle request %r (extra=%r)",
            cmd, extra_data,
        )
        return protocol.ToplevelCommand.ERROR, None


class Scheduler:
    DEFAULT_PRUNE_SCHEDULE = {
        "interval_step": 1,
        "interval_unit": "day",
    }

    def __init__(self, loop, socket_path):
        super().__init__()
        self.loop = loop
        self.logger = logging.getLogger("cuteborg.scheduler")
        self.socket_path = socket_path
        self.backend = cuteborg.subprocess_backend.LocalSubprocessBackend(
            logger=self.logger.getChild("backend")
        )

        self.wctsleep = None

        self._default_prune_schedule = config.ScheduleConfig.from_raw(
            self.DEFAULT_PRUNE_SCHEDULE
        )

        self._endpoint = None
        self._wakeup_event = None
        self._stop_event = None

        self._last_runs = {}
        self._jobs = {}
        self._job_tasks = {}
        self._job_blockers = {}
        self._job_progress = {}
        self._job_errors = {}
        self._sleep_tasks = {}
        self._repository_locks = {}
        self._deferred_jobs = set()

        self._load_state()

        self._rls_pending = False
        self._shutting_down = False
        self._reload_and_reschedule(defer=True)

    def _save_state(self):
        cfg_path = pathlib.Path(
            xdg.BaseDirectory.xdg_config_home
        ) / "cuteborg" / "scheduler-state.toml"

        self.logger.debug("saving state to %s", cfg_path)

        state = {
            "last_runs": [
                {
                    "key": list(key),
                    "timestamp": ts,
                }
                for key, ts in self._last_runs.items()
            ]
        }

        config.toml_to_file(state, cfg_path)

    def _load_state(self):
        cfg_path = pathlib.Path(
            xdg.BaseDirectory.xdg_config_home
        ) / "cuteborg" / "scheduler-state.toml"

        self.logger.debug("loading state from %s", cfg_path)
        try:
            with cfg_path.open("r") as f:
                state = toml.load(f)
        except OSError as exc:
            self.logger.warning("failed to load state: %s")
            return

        self._last_runs = {
            tuple(item["key"]): item["timestamp"]
            for item in state.get("last_runs", [])
        }

    def _create_protocol(self):
        self.logger.debug("new connection")
        proto = protocol.ControlProtocol(
            self.logger.getChild("control"),
            self._endpoint.handle_request,
        )
        return proto

    def _job_main_task_done(self, job_obj, task):
        try:
            task.result()
        except asyncio.CancelledError:
            try:
                ex_job_obj, ex_main_task = self._jobs[job_obj.key]
            except KeyError:
                # task was cancelled because the job was erased
                pass
            else:
                if ex_job_obj is job_obj and not self._shutting_down:
                    # task was cancelled for unknown reasons
                    self.logger.warning(
                        "unexpected cancellation of main task for job %r",
                        job_obj,
                    )
                # task was cancelled because it was replaced
        except Exception as exc:
            self.logger.exception(
                "main task for job %r failed",
                job_obj
            )
            self.set_job_error(job_obj.key, str(exc))
        else:
            self.logger.warning(
                "main task for job %r terminated",
                job_obj
            )

    def _job_task_done(self, job_obj, start_time, task):
        try:
            task.result()
        except asyncio.CancelledError:
            if not self._shutting_down:
                self.logger.warning(
                    "unexpected cancellation of task for job %r",
                    job_obj,
                )
        except:
            self.logger.exception(
                "task for job %r failed",
                job_obj
            )
        else:
            # clear errors
            self._job_errors.pop(job_obj.key, None)

        if self._job_tasks[job_obj.key] != task:
            self.logger.warning(
                "wtf? inconsistent internal state: "
                "self._job_tasks[job_obj.key] != task ... "
                "unsure what to do, continuing for now"
            )

        del self._job_tasks[job_obj.key]

        try:
            del self._job_progress[job_obj.key]
        except KeyError:
            pass

        self.logger.debug(
            "recording last run of %r as %s",
            job_obj,
            start_time
        )
        self._last_runs[job_obj.key] = start_time

        job_obj, main_task = self._jobs[job_obj.key]
        if main_task is None:
            self.logger.debug(
                "starting main task for %r now after it has been deferred "
                "earlier",
                job_obj
            )
            main_task = asyncio.ensure_future(
                job_obj.run()
            )
            self._jobs[job_obj.key] = job_obj, main_task

        self._endpoint.notify_poll()

    def _add_job(self, job_obj, defer):
        key = job_obj.key

        try:
            job_obj, main_task = self._jobs[key]
        except KeyError:
            if defer:
                self.logger.debug("deferring startup of task for job %r",
                                  job_obj)
                self._deferred_jobs.add(key)
                main_task = None
            else:
                self.logger.debug("starting main task for job %r", job_obj)
                main_task = asyncio.ensure_future(job_obj.run())
                main_task.add_done_callback(
                    functools.partial(
                        self._job_main_task_done,
                        job_obj,
                    )
                )
        else:
            self.logger.debug(
                "deferring start of main task for job %r, as a task "
                "is currently running",
                job_obj
            )
            main_task = None

        self._jobs[key] = (
            job_obj, main_task
        )

    def _reload_and_reschedule(self, defer=False):
        self._rls_pending = False
        self.logger.debug("reloading configuration")
        self.config = config.Config.from_raw(config.load_config())

        self.logger.debug(
            "found %d repositories with %d jobs in total",
            len(self.config.repositories),
            len(self.config.jobs),
        )

        for job_obj, job_main_task in self._jobs.values():
            if job_main_task is not None:
                self.logger.debug("stopping main task of job %s", job_obj.key)
                job_main_task.cancel()

        self._jobs.clear()
        self._deferred_jobs.clear()
        self._job_errors.clear()

        new_repositories = set(self.config.repositories)
        old_repositories = set(self._repository_locks)

        self.logger.debug("old repositories = %r, new repositories = %r",
                          old_repositories, new_repositories)

        if not defer:
            for id_ in (new_repositories - old_repositories):
                # create locks for added repositories
                self.logger.debug("repository %r: preparing", id_)
                self._repository_locks[id_] = asyncio.Lock(loop=self.loop)

            for id_ in (old_repositories - new_repositories):
                self.logger.debug("repository %r removed -- tearing down", id_)
                del self._repository_locks[id_]

        # At this point, last_run information needs to be available

        for id_, repository in self.config.repositories.items():
            if repository.prune is not None:
                self.logger.debug(
                    "repository %r: setting up prune job",
                    id_
                )

                instance_logger = self.logger.getChild(
                    "prune",
                ).getChild(
                    str(id_)
                )

                job_obj = jobs.PruneRepository(
                    instance_logger,
                    self,
                    repository.prune,
                    repository,
                    repository.prune.schedule or self._default_prune_schedule,
                )

                self._add_job(job_obj, defer)

        for name, job in self.config.jobs.items():
            for repository in job.repositories:
                self.logger.debug("job %r, repository %r: preparing",
                                  name, repository.id_)
                instance_logger = self.logger.getChild(
                    "create_archive"
                ).getChild(
                    job.name
                ).getChild(
                    str(repository.id_)
                )

                job_obj = jobs.CreateArchive(
                    instance_logger,
                    self,
                    job,
                    repository,
                    job.schedule or self.config.schedule,
                )

                self._add_job(job_obj, defer)

        self.logger.debug("repository_locks = %r", self._repository_locks)
        self.logger.debug("jobs = %r", self._jobs)
        self.logger.debug("job_tasks = %r", self._job_tasks)
        self.logger.debug("max_slack = %r", self.config.max_slack)
        self.logger.debug("poll_interval = %r", self.config.poll_interval)

        if self.wctsleep is not None:
            self.wctsleep.max_slack = self.config.max_slack
        if self._wakeup_event is not None:
            self._wakeup_event.set()
        if self._endpoint is not None:
            self._endpoint.notify_poll()

    def _scheduled_reload_and_reschedule(self):
        if not self._rls_pending:
            return
        self._reload_and_reschedule()

    def _trigger_reload_and_reschedule(self):
        if self._rls_pending:
            return

        self._rls_pending = True
        self.loop.call_later(
            1,
            self._scheduled_reload_and_reschedule
        )

    def _trigger_exit(self):
        self._stop_event.set()

    def _signalled_reload_and_reschedule(self):
        self.logger.info(
            "reload and reschedule requested via SIGUSR1/SIGHUP -- "
            "scheduling reload"
        )
        self._trigger_reload_and_reschedule()

    @asyncio.coroutine
    def _main_loop(self):
        for key in self._deferred_jobs:
            try:
                job_obj, main_task = self._jobs.pop(key)
            except KeyError:
                continue

            if main_task is None:
                self._add_job(job_obj, False)
            else:
                self._jobs[key] = job_obj, main_task

        self._deferred_jobs.clear()

        with contextlib.ExitStack() as stack:
            stop_future = stack.enter_context(
                cancelling(
                    asyncio.async(
                        self._stop_event.wait(),
                        loop=self.loop)
                )
            )

            self._wakeup_event.clear()
            wakeup_future = stack.enter_context(
                cancelling(
                    asyncio.async(
                        self._wakeup_event.wait(),
                        loop=self.loop)
                )
            )

            done, pending = yield from asyncio.wait(
                [
                    stop_future,
                    wakeup_future,
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )

            if stop_future in done:
                self.logger.info("received stop signal (SIGTERM or SIGINT)")
                return False

            return True

    @asyncio.coroutine
    def check_server_socket(self):
        try:
            endpoint = yield from asyncio.wait_for(
                protocol.test_and_get_socket(
                    self.loop,
                    self.logger,
                    self.socket_path,
                ),
                timeout=5
            )
        except (FileNotFoundError):
            pass
        except ConnectionError:
            self.logger.error(
                "socket %r exists, but it doesn’t conform to scheduler "
                "protocol. aborting.",
                str(self.socket_path)
            )
            return 1
        except asyncio.TimeoutError:
            self.logger.error(
                "socket %r exists, but no reply received there in time."
                " is there a running scheduler which is stuck?",
                str(self.socket_path)
            )
            return 1
        else:
            self.logger.error(
                "socket %r exists, and a scheduler is replying there! "
                "aborting.",
                str(self.socket_path),
            )
            endpoint.close()
            return 1

        try:
            self.socket_path.parent.mkdir(mode=0o700, parents=True)
        except FileExistsError:
            if not self.socket_path.parent.is_dir():
                raise

        if self.socket_path.exists():
            self.logger.error(
                "socket %s exists! please remove it.",
                self.socket_path,
            )
            return 1

    def create_server_socket(self):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM, 0)
        sock.bind(str(self.socket_path))
        return sock

    @asyncio.coroutine
    def bind_server_socket(self, sock):
        return (yield from self.loop.create_unix_server(
            self._create_protocol,
            sock=sock,
        ))

    @asyncio.coroutine
    def main(self, sock=None, loop=None):
        if sock is None:
            yield from self.check_server_socket()
            sock = self.create_server_socket()

        if loop is not None and loop != self.loop:
            self.loop = loop

        self._endpoint = ServerProtocolEndpoint(
            self.logger.getChild("control"),
            self
        )
        self._wakeup_event = asyncio.Event(loop=loop)
        self._stop_event = asyncio.Event(loop=loop)
        self.wctsleep = wctsleep.WallClockTimeSleepImpl()
        self._repository_locks.clear()
        self._reload_and_reschedule()

        self.loop.add_signal_handler(signal.SIGINT, self._stop_event.set)
        self.loop.add_signal_handler(signal.SIGTERM, self._stop_event.set)

        self.loop.add_signal_handler(
            signal.SIGHUP,
            self._signalled_reload_and_reschedule,
        )
        self.loop.add_signal_handler(
            signal.SIGUSR1,
            self._signalled_reload_and_reschedule,
        )

        server = yield from self.bind_server_socket(sock)

        self.logger.info("server started up")

        try:
            # main_loop returns True while it should keep looping
            # and False otherwise
            while (yield from self._main_loop()):
                pass
        finally:
            self.logger.info("saving state")
            self._save_state()
            self.logger.info("shutting down server")
            server.close()
            self._stop_event.clear()
            self.logger.warning(
                "stopping server and tasks -- "
                "if it hangs, send SIGINT/SIGTERM again"
            )

            self._shutting_down = True

            for job_task in self._job_tasks.values():
                job_task.cancel()

            for _, job_main_task in self._jobs.values():
                job_main_task.cancel()

            other_futures = list(self._job_tasks.values())
            other_futures.extend(
                job_main_task
                for _, job_main_task in self._jobs.values()
            )
            other_futures.append(
                server.wait_closed()
            )
            other_futures.append(
                self.wctsleep.cancel_and_wait()
            )

            other_futures = [
                asyncio.ensure_future(fut)
                for fut in other_futures
            ]

            other_futures_fut = asyncio.ensure_future(
                asyncio.wait(
                    other_futures,
                ),
            )

            def debug_done(fut):
                self.logger.debug("shutdown: %r finished", fut)

            for fut in other_futures:
                fut.add_done_callback(debug_done)
                self.logger.debug("waiting for: %r", fut)

            self.logger.warning(
                "waiting for server and tasks -- "
                "if it hangs, send SIGINT/SIGTERM again"
            )

            stop_fut = asyncio.ensure_future(self._stop_event.wait())

            done, pending = yield from asyncio.wait(
                [
                    stop_fut,
                    other_futures_fut,
                ],
                return_when=asyncio.FIRST_COMPLETED
            )

            for fut in pending:
                if fut != stop_fut:
                    self.logger.warning("not finished: %r", fut)
                fut.cancel()

            self.logger.debug(
                "shutdown completed"
            )

    def get_last_run(self, key):
        return self._last_runs.get(key)

    @asyncio.coroutine
    def sleep_until(self, job_obj, dt):
        with note_job_block(
                self._job_blockers,
                job_obj.key,
                (
                    "sleep_until",
                    (dt, ),
                ),
                self._endpoint.notify_poll):
            wait_task = asyncio.ensure_future(
                self.wctsleep.sleep_until(dt)
            )
            self._sleep_tasks[job_obj.key] = wait_task, False
            try:
                yield from wait_task
            except asyncio.CancelledError:
                _, awoken = self._sleep_tasks.get(job_obj.key, (None, False))
                if awoken:
                    return
                raise
            finally:
                self._sleep_tasks.pop(job_obj.key, None)

    def execute_job(self, job_obj, **kwargs):
        key = job_obj.key
        task = asyncio.ensure_future(job_obj.exec(**kwargs))
        self._job_tasks[key] = task
        task.add_done_callback(
            functools.partial(
                self._job_task_done,
                job_obj,
                datetime.utcnow()
            )
        )
        return asyncio.ensure_future(asyncio.shield(task))

    def _force_archive(self, job, only=None):
        if only is not None:
            keys = [
                ("create_archive", job, repo)
                for repo in only
            ]
        else:
            keys = [
                key
                for key in self._sleep_tasks.keys()
                if key[0] == "create_archive" and key[1] == job
            ]

        tasks = []
        for key in keys:
            sleeper, _ = self._sleep_tasks[key]
            self._sleep_tasks[key] = sleeper, True
            sleeper.cancel()

    def _force_prune(self, repositories):
        keys = [
            ("prune", repo)
            for repo in repositories
        ]

        tasks = []
        for key in keys:
            sleeper, _ = self._sleep_tasks[key]
            self._sleep_tasks[key] = sleeper, True
            sleeper.cancel()

    @asyncio.coroutine
    def _wait_for_remote_repository(self, logger, job_key, remote_repo,
                                    update_status):
        context = cuteborg.backend.Context()
        remote_repo.setup_context(context)

        path = remote_repo.make_repository_path()
        while True:
            try:
                yield from self.backend.ping(path, context)
            except cuteborg.backend.RepositoryUnreachable as exc:
                update_status(str(exc))
            else:
                break

            yield from asyncio.sleep(self.config.poll_interval)

        return path

    @asyncio.coroutine
    def _wait_for_removable_device(self, logger, job_key, local_repo):
        problem_known = None
        while True:
            try:
                mount_path = yield from devices.wait_for_mounted(
                    logger,
                    local_repo.removable_device_uuid,
                    self.config.poll_interval,
                    crypto_passphrase=local_repo.crypto_passphrase
                )
                break
            except devices.EncryptedDeviceWithoutPassphrase as exc:
                if problem_known is not type(exc):
                    self.set_job_error(job_key, str(exc))
                problem_known = type(exc)

            yield from asyncio.sleep(self.config.poll_interval)

        # clear errors
        self._job_errors.pop(job_key, None)

        logger.debug(
            "device is mounted at %r",
            mount_path
        )
        return mount_path + local_repo.make_repository_path()

    @asyncio.coroutine
    def wait_for_repository_storage(self, job_obj, repository_cfg):
        if isinstance(repository_cfg, config.RemoteRepositoryConfig):
            with contextlib.ExitStack() as stack:
                has_block = False

                def update_status(msg):
                    nonlocal has_block
                    argv = (
                        "network_host",
                        (repository_cfg.host, msg),
                    )
                    if not has_block:
                        stack.enter_context(
                            note_job_block(
                                self._job_blockers,
                                job_obj.key,
                                (
                                    "network_host",
                                    (repository_cfg.host, msg),
                                ),
                                self._endpoint.notify_poll)
                        )
                        has_block = True
                    else:
                        self._job_blockers[job_obj.key] = argv
                        self._endpoint.notify_poll()

                return (yield from self._wait_for_remote_repository(
                    job_obj.logger,
                    job_obj.key,
                    repository_cfg,
                    update_status,
                ))
        elif isinstance(repository_cfg, config.LocalRepositoryConfig):
            if repository_cfg.removable_device_uuid is not None:
                with note_job_block(
                        self._job_blockers,
                        job_obj.key,
                        (
                            "removable_device",
                            (repository_cfg.removable_device_uuid, ),
                        ),
                        self._endpoint.notify_poll):
                    return (yield from self._wait_for_removable_device(
                        job_obj.logger,
                        job_obj.key,
                        repository_cfg
                    ))
            else:
                return repository_cfg.make_repository_path()

    @asyncio.coroutine
    def lock_repository(self, job_obj, id_):
        with note_job_block(
                self._job_blockers,
                job_obj.key,
                (
                    "waiting_for_repository_lock",
                    (id_,)
                ),
                self._endpoint.notify_poll):
            return (yield from self._repository_locks[id_])

    def get_job_status(self):
        for key, (job_obj, job_main_task) in self._jobs.items():
            yield (
                key,
                job_obj,
                self._job_tasks.get(key),
                self._job_blockers.get(key),
                self._job_progress.get(key),
                self._job_errors.get(key),
                self._last_runs.get(key),
            )

    def set_job_progress(self, job_obj, progress):
        self._job_progress[job_obj.key] = progress
        self._endpoint.notify_poll()

    def set_job_error(self, job_key, message, timestamp=None):
        if message is None:
            try:
                del self._job_errors[job_key]
            except KeyError:
                pass
            else:
                self._endpoint.notify_poll()
            return

        timestamp = timestamp or datetime.utcnow()
        self._job_errors[job_key] = (
            message,
            timestamp,
        )
        self._endpoint.notify_poll()



########################################################################
# File name: __init__.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import enum
import itertools
import pathlib
import signal
import subprocess

from datetime import datetime

import babel.dates

import PyQt5.Qt as Qt

import cuteborg.scheduler.protocol

from qtborg import utils

from .ui_main import Ui_StatusWindow


SortRole = Qt.Qt.UserRole


@asyncio.coroutine
def try_start():
    proc = yield from asyncio.create_subprocess_exec(
        "cuteborg-scheduler",
        "-vv",
        "run",
        "-d",
        stdin=subprocess.DEVNULL,
    )

    yield from proc.wait()

    return proc.returncode == 0


@asyncio.coroutine
def try_stop():
    proc = yield from asyncio.create_subprocess_exec(
        "cuteborg-scheduler",
        "-vvv",
        "stop",
        stdin=subprocess.DEVNULL,
    )

    yield from proc.wait()

    return proc.returncode == 0


def _load_pixmap(name):
    for p in __path__:
        filename = pathlib.Path(p) / "data" / name
        pixmap = Qt.QPixmap(str(filename))
        if pixmap.isNull():
            continue
        return pixmap
    raise FileNotFoundError("icon {!r} not found".format(name))


def _load_icon(name):
    return Qt.QIcon(_load_pixmap(name))


def _compose_icon(bottom_left=None, top_right=None):
    base = _load_pixmap("base.png")
    if bottom_left is None and top_right is None:
        return Qt.QIcon(base)

    if bottom_left is not None:
        bottom_left = _load_pixmap(bottom_left).toImage()
    if top_right is not None:
        top_right = _load_pixmap(top_right).toImage()

    base = base.toImage()
    painter = Qt.QPainter()
    painter.begin(base)

    if bottom_left is not None:
        y = base.height() - bottom_left.height()
        painter.drawImage(
            Qt.QPoint(0, y),
            bottom_left
        )

    if top_right is not None:
        x = base.width() - top_right.width()
        painter.drawImage(
            Qt.QPoint(x, 0),
            top_right
        )

    painter.end()

    return Qt.QIcon(Qt.QPixmap.fromImage(base))


CREATE_ARCHIVE_BASE_TEMPLATE = (
    "<tr><td>archive</td>"
    "<td>{job}</td>"
    "<td>{repo[name]}</td>"
    "<td>{status}</td></tr>"
)

CREATE_ARCHIVE_PROGRESS_TEMPLATE = (
    "{progress[uncompressed]} archived"
)

PRUNE_RUNNING_TEMPLATE = (
    "<tr><td>prune</td>"
    "<td>{progress[job]}</td>"
    "<td>{repo[name]}</td>"
    "<td>in progress</td></tr>"
)

PRUNE_NONRUNNING_TEMPLATE = (
    "<tr><td>prune</td>"
    "<td>—</td>"
    "<td>{repo[name]}</td>"
    "<td>{status}</td></tr>"
)


def format_job_state_table(job, now, parts=None):
    parts = parts or []

    if "progress" in job:
        progress = job["progress"]
        if progress.get("synced", 1) < 1:
            parts.append("synchronizing")
        else:
            parts.append(CREATE_ARCHIVE_PROGRESS_TEMPLATE.format(**job))

    if "blocking" in job:
        blocking_info = job["blocking"]
        if blocking_info["reason"] == "sleep_until":
            parts.append(
                babel.dates.format_timedelta(
                    blocking_info["struct_info"]["wakeup_at"] - now,
                    add_direction=True,
                )
            )

        elif blocking_info["reason"] == "waiting_for_repository_lock":
            parts.append(
                "waiting for other job to finish",
            )

        elif blocking_info["reason"] == "removable_device":
            parts.append(
                "waiting for removable device",
            )

        elif blocking_info["reason"] == "network_host":
            parts.append(
                "waiting for network connectivity"
            )

        else:
            parts.append(blocking_info["reason"])

    if "error" in job:
        parts.append(
            "<strong>Error:</strong> {}".format(
                job["error"]["message"]
            )
        )

    return "<br/>".join(parts)


def format_job_state_view(job):
    parts = []

    if "blocking" in job:
        blocking_info = job["blocking"]
        if blocking_info["reason"] == "sleep_until":
            # shown at other place in table
            pass

        elif blocking_info["reason"] == "waiting_for_repository_lock":
            parts.append(
                "waiting for other job to finish",
            )

        elif blocking_info["reason"] == "removable_device":
            parts.append(
                "waiting for removable device",
            )

        elif blocking_info["reason"] == "network_host":
            parts.append(
                "waiting for network connectivity"
            )

        else:
            parts.append(blocking_info["reason"])

    if "error" in job:
        parts.append(
            "error: {}".format(
                job["error"]["message"]
            )
        )

    return "; ".join(parts)


class JobsModel(Qt.QAbstractTableModel):
    COLUMN_TYPE = 0
    COLUMN_JOB = 1
    COLUMN_REPOSITORY = 2
    COLUMN_STATE = 3
    COLUMN_LAST_RUN = 4
    COLUMN_NEXT_RUN = 5
    COLUMN_PROGRESS = 6
    COLUMN_COUNT = 7

    REFERENCE_TIME = datetime(1970, 1, 1, 1, 0)

    def __init__(self):
        super().__init__()
        self.jobs = []
        self.known_keys = set()

    def update_jobs(self, jobs):
        update_tmp = {}
        abandoned_keys = set(self.known_keys)

        for job in jobs:
            key = (job["type"], ) + tuple(job["raw_args"])

            abandoned_keys.discard(key)

            if key in self.known_keys:
                update_tmp[key] = job
                continue
            self.known_keys.add(key)

            self.beginInsertRows(Qt.QModelIndex(),
                                 len(self.jobs),
                                 len(self.jobs))
            self.jobs.append(job)
            self.endInsertRows()

        indices_to_delete = []

        for i, job in enumerate(self.jobs):
            key = (job["type"], ) + tuple(job["raw_args"])

            if key in abandoned_keys:
                indices_to_delete.append(i)
                self.known_keys.discard(key)
                continue

            try:
                update = update_tmp.pop(key)
            except KeyError:
                continue

            job.clear()
            job.update(update)

            self.dataChanged.emit(
                self.index(i, 0),
                self.index(i, self.COLUMN_COUNT),
            )

        for index in reversed(indices_to_delete):
            self.beginRemoveRows(Qt.QModelIndex(), index, index)
            del self.jobs[index]
            self.endRemoveRows()

    def rowCount(self, parent):
        if parent.isValid():
            return 0
        return len(self.jobs)

    def columnCount(self, parent):
        return self.COLUMN_COUNT

    def index(self, row, column, parent=None):
        return self.createIndex(
            row,
            column,
            self.jobs[row],
        )

    def headerData(self, section, orientation, role):
        if orientation == Qt.Qt.Horizontal:
            if role == Qt.Qt.DisplayRole:
                return [
                    "Type",
                    "Job",
                    "Repository",
                    "State",
                    "Last run",
                    "Next run",
                    "Progress",
                ][section]
            elif role == Qt.Qt.InitialSortOrderRole:
                return [
                    Qt.Qt.AscendingOrder,
                    Qt.Qt.AscendingOrder,
                    Qt.Qt.AscendingOrder,
                    Qt.Qt.AscendingOrder,
                    Qt.Qt.AscendingOrder,
                    Qt.Qt.DescendingOrder,
                    Qt.Qt.AscendingOrder,
                ][section]

    def data(self, index, role):
        if role != Qt.Qt.DisplayRole and role != SortRole:
            return

        if not index.isValid():
            return

        job = index.internalPointer()

        if index.column() == self.COLUMN_TYPE:
            if job["type"] == "create_archive":
                return "archive"
            elif job["type"] == "prune":
                return "prune"
            else:
                return job["type"]

        elif index.column() == self.COLUMN_JOB:
            try:
                return job["args"]["job"]
            except KeyError:
                if job["type"] == "prune":
                    try:
                        return job["progress"]["job"]
                    except KeyError:
                        pass

        elif index.column() == self.COLUMN_REPOSITORY:
            try:
                return job["args"]["repo"]["name"]
            except KeyError:
                pass

        elif index.column() == self.COLUMN_STATE:
            if job["running"] and not job.get("blocking"):
                state = "running"
            elif job.get("error"):
                state = "failed"
            elif job.get("blocking"):
                if job["blocking"]["reason"] == "sleep_until":
                    state = "scheduled"
                else:
                    state = "waiting"
            else:
                state = "rescheduling"

            return state

        elif index.column() == self.COLUMN_LAST_RUN:
            try:
                last_run = job["last_run"]
            except KeyError:
                if role == SortRole:
                    return float("-inf")
                return "never"
            if role == SortRole:
                if last_run is None:
                    last_run = datetime.utcnow()
                return (last_run - self.REFERENCE_TIME).total_seconds()
            elif role == Qt.Qt.ToolTipRole:
                return str(last_run)
            else:
                return babel.dates.format_timedelta(
                    last_run - datetime.utcnow(),
                    add_direction=True,
                    granularity='minute',
                )

        elif index.column() == self.COLUMN_NEXT_RUN:
            try:
                wakeup_at = job["blocking"]["struct_info"]["wakeup_at"]
            except KeyError:
                if role == SortRole:
                    return float("-inf")
                return "now"
            if role == SortRole:
                return (wakeup_at - self.REFERENCE_TIME).total_seconds()
            elif role == Qt.Qt.ToolTipRole:
                return str(wakeup_at)
            else:
                return babel.dates.format_timedelta(
                    wakeup_at - datetime.utcnow(),
                    granularity='minute',
                    add_direction=True,
                )

        elif index.column() == self.COLUMN_PROGRESS:
            if job["type"] == "create_archive" and "progress" in job:
                progress = job["progress"]
                if progress["synced"] < 1:
                    return "synchronizing ({:.0f}%)".format(
                        progress["synced"] * 100,
                    )
                return "archived {} from {} files".format(
                    job["progress"]["uncompressed"],
                    job["progress"]["nfiles"],
                )
            elif job["type"] == "prune" and "progress" in job:
                return "pruning archives from job {!r}".format(
                    job["progress"]["job"]
                )
            else:
                return format_job_state_view(job)

    def flags(self, index):
        return Qt.Qt.ItemIsEnabled | Qt.Qt.ItemIsSelectable


class ErrorCondition(enum.Enum):
    SCHEDULER_NOT_RUNNING = "The scheduler is not running!"
    REMOVABLE_DEVICE_MISSING = \
        "A removable device ({}) must be inserted for backup jobs to continue."
    OTHER_ERROR = "Error: {}"
    NETWORK_HOST = \
        "Cannot access borg server at {!r}{}."


class Main(Qt.QMainWindow):
    def __init__(self, loop, logger, socket_path):
        super().__init__()
        self.logger = logger

        self.ui = Ui_StatusWindow()
        self.ui.setupUi(self)

        self.trayicon = Qt.QSystemTrayIcon()
        self.trayicon.activated.connect(
            self._tray_icon_activated
        )
        self.trayicon.installEventFilter(self)

        self.icons = {
            (colour, flag): _compose_icon(
                flag and "bullet_{}.png".format(flag),
                colour and "bullet_{}.png".format(colour),
            )
            for colour, flag in itertools.product(
                (None, "blue", "orange", "red"),
                (None, "error"))
        }

        self._connected = False
        self._stop = False
        self._quit = False
        self._jobs = []
        self._error_conditions = set()
        self._curr_icon = None, "error"
        self.trayicon.setIcon(
            self.icons[self._curr_icon],
        )
        self.socket_path = socket_path
        self.loop = loop

        self.stop_event = asyncio.Event()

        self.blink_task = None

        self.jobs_model = JobsModel()
        proxy_model = Qt.QSortFilterProxyModel(self.ui.jobs_view)
        proxy_model.setSourceModel(self.jobs_model)
        proxy_model.setSortRole(SortRole)
        proxy_model.setSortCaseSensitivity(Qt.Qt.CaseInsensitive)
        proxy_model.setSortLocaleAware(True)
        proxy_model.sort(
            JobsModel.COLUMN_NEXT_RUN,
            Qt.Qt.DescendingOrder,
        )

        self.ui.jobs_view.setModel(proxy_model)
        self.ui.jobs_view.horizontalHeader().setSectionResizeMode(
            Qt.QHeaderView.ResizeToContents,
        )
        self.ui.jobs_view.horizontalHeader().setSortIndicator(
            JobsModel.COLUMN_NEXT_RUN,
            Qt.Qt.DescendingOrder,
        )

        self.ui.action_close.triggered.connect(
            self._action_close,
        )

        self.ui.action_quit.triggered.connect(
            self._action_quit,
        )

        self.ui.action_quit_and_stop.triggered.connect(
            self._action_quit_and_stop,
        )

        signal.signal(
            signal.SIGINT,
            signal.SIG_DFL,
        )

        signal.signal(
            signal.SIGTERM,
            signal.SIG_DFL,
        )

    def _action_close(self):
        self.hide()

    def _action_quit(self, _):
        result = Qt.QMessageBox.warning(
            self,
            "Exit CuteBorg Status",
            "Exiting CuteBorg status will not stop the CuteBorg "
            "scheduler which runs your backups, but will remove "
            "the indicator icon which shows you the status.",
            Qt.QMessageBox.Ok | Qt.QMessageBox.Cancel,
        )

        self._quit = True
        if result == Qt.QMessageBox.Ok:
            self.stop_event.set()

    def _action_quit_and_stop(self, _):
        result = Qt.QMessageBox.warning(
            self,
            "Exit CuteBorg Status and stop CuteBorg",
            "If you proceed, your backups will not continue to be "
            "run until you re-start CuteBorg manually.",
            Qt.QMessageBox.Ok | Qt.QMessageBox.Cancel,
        )

        self._quit = True
        self._stop = True
        if result == Qt.QMessageBox.Ok:
            self.stop_event.set()

    def _tray_icon_activated(self, reason):
        if reason == Qt.QSystemTrayIcon.Trigger:
            self.show()
            self.raise_()
            return

    @asyncio.coroutine
    def _blink_task(self):
        try:
            while True:
                yield from asyncio.sleep(1.5)
                self.trayicon.setIcon(self.icons[None, None])
                yield from asyncio.sleep(0.5)
                self.trayicon.setIcon(self.icons[self._curr_icon])
        except asyncio.CancelledError:
            self.trayicon.setIcon(self.icons[self._curr_icon])

    def _update_jobs(self, jobs):
        self._jobs = jobs
        self.jobs_model.update_jobs(jobs)

    def _set_error_condition(self, condition, data=(), *,
                             suppress_notify=False):
        key = condition, data
        if key in self._error_conditions:
            return

        self._error_conditions.add(key)
        self._update_status_label()

        if not suppress_notify and Qt.QSystemTrayIcon.supportsMessages():
            self.trayicon.showMessage(
                "CuteBorg error",
                condition.value.format(*data),
                Qt.QSystemTrayIcon.Warning,
            )

    def _update_status_label(self):
        if self._error_conditions:
            self.ui.status_box.show()
            self.ui.status_icon.setPixmap(
                Qt.QApplication.style().standardIcon(
                    Qt.QStyle.SP_MessageBoxWarning,
                ).pixmap(48)
                # self.icons[None, "error"].pixmap(48),
            )
            text = []
            for condition, data in self._error_conditions:
                text.append(
                    "<li>{}</li>".format(
                        condition.value.format(*data)
                    )
                )

            self.ui.status_label.setText(
                "<h6>Problems:</h6><ul>{}</ul>".format("".join(text))
            )
        else:
            self.ui.status_box.hide()

    def _clear_error_condition(self, condition, data=None):
        self._error_conditions.discard((condition, data))
        self._update_status_label()

    def _update_error_conditions_from_jobs(self):
        current_conditions = set()

        for job in self._jobs:
            try:
                error_info = job["error"]
            except KeyError:
                pass
            else:
                current_conditions.add(
                    (ErrorCondition.OTHER_ERROR,
                     (error_info["message"],))
                )
                continue

            try:
                blocking_info = job["blocking"]
            except KeyError:
                pass
            else:
                if blocking_info["reason"] == "removable_device":
                    current_conditions.add(
                        (ErrorCondition.REMOVABLE_DEVICE_MISSING,
                         (blocking_info["struct_info"]["device_uuid"],))
                    )
                elif blocking_info["reason"] == "network_host":
                    extra = blocking_info["struct_info"].get("extra")
                    if extra is not None:
                        extra = " ({})".format(extra)
                    current_conditions.add(
                        (ErrorCondition.NETWORK_HOST,
                         (blocking_info["struct_info"]["host"],
                          extra))
                    )

        for key in set(self._error_conditions):
            if key not in current_conditions:
                self._clear_error_condition(*key)

        for key in current_conditions:
            self._set_error_condition(*key)

    def _generate_tooltip_from_jobs(self):
        job_table = []
        nrunning = 0
        now = datetime.utcnow()

        for job in self._jobs:
            if job["running"]:
                nrunning += 1

            if job["type"] == "create_archive":
                job_table.append(
                    CREATE_ARCHIVE_BASE_TEMPLATE.format(
                        job=job["args"]["job"],
                        repo=job["args"]["repo"],
                        status=(
                            format_job_state_table(job, now)
                        ),
                    )
                )

            elif job["type"] == "prune":
                if "progress" in job:
                    job_table.append(
                        PRUNE_RUNNING_TEMPLATE.format(
                            repo=job["args"]["repo"],
                            progress=job["progress"],
                        )
                    )
                else:
                    job_table.append(
                        PRUNE_NONRUNNING_TEMPLATE.format(
                            repo=job["args"]["repo"],
                            status=format_job_state_table(job, now),
                        )
                    )

        return "<p>running jobs: {}</p>{}".format(
            nrunning,
            ("<table><tr><th>Task</th><th>Job name</th><th>Repository</th><th>Status</th></tr>{}</table>".format("".join(job_table)))
            if job_table else ""
        )

    def eventFilter(self, object_, ev):
        if object_ == self.trayicon:
            if ev.type() == Qt.QEvent.ToolTip:
                if self._connected:
                    Qt.QToolTip.showText(
                        ev.globalPos(),
                        self._generate_tooltip_from_jobs(),
                    )
                else:
                    Qt.QToolTip.showText(
                        ev.globalPos(),
                        "<strong>Scheduler not running!</strong>"
                    )
                return True
            return False
        else:
            return False

    def _update_trayicon_from_jobs(self):
        nrunning = False
        any_failed = False
        any_warning = False

        for job in self._jobs:
            if job["running"]:
                nrunning += 1

            any_failed = (
                any_failed or
                "error" in job and "blocking" not in job
            )

            any_warning = (
                any_warning or
                "error" in job or
                job.get("blocking", {}).get("reason") == "removable_device"
            )

        if any_failed:
            self._curr_icon = "red", None
        elif any_warning:
            self._curr_icon = "orange", None
        else:
            self._curr_icon = "blue", None

        if nrunning and self.blink_task is None:
            self.trayicon.setIcon(self.icons[self._curr_icon])

            self.blink_task = asyncio.ensure_future(
                self._blink_task()
            )

            def clear_task(fut):
                self.blink_task = None
                fut.result()

            self.blink_task.add_done_callback(
                clear_task
            )
        elif not nrunning and self.blink_task is not None:
            self.blink_task.cancel()
            self.blink_task = None

        if not nrunning:
            self.trayicon.setIcon(self.icons[self._curr_icon])

    @asyncio.coroutine
    def _get_connection(self):
        while True:
            try:
                endpoint = yield from asyncio.wait_for(
                    cuteborg.scheduler.protocol.test_and_get_socket(
                        self.loop,
                        self.logger,
                        self.socket_path,
                    ),
                    timeout=5
                )
                break
            except (asyncio.TimeoutError,
                    FileNotFoundError,
                    ConnectionError) as exc:
                msg = str(exc)
                if isinstance(exc, asyncio.TimeoutError):
                    msg = "timeout"
                self.logger.error("failed to connect to scheduler at %r: %s",
                                  str(self.socket_path),
                                  msg)
            self._update_jobs({})
            self._set_error_condition(
                ErrorCondition.SCHEDULER_NOT_RUNNING,
                suppress_notify=not self._connected
            )
            yield from asyncio.sleep(10)

        self._connected = True

        response, data = yield from endpoint.request(
            cuteborg.scheduler.protocol.ToplevelCommand.STATUS,
        )

        self._update_jobs(data["jobs"])

        return endpoint

    @asyncio.coroutine
    def _update_loop(self, stop_future):
        self._curr_icon = None, "error"
        self._set_error_condition(
            ErrorCondition.SCHEDULER_NOT_RUNNING,
            suppress_notify=not self._connected
        )
        self._connected = False
        self.trayicon.setIcon(self.icons[self._curr_icon])

        self.logger.debug("attempting connection ...")

        conn_future = asyncio.ensure_future(self._get_connection())
        done, pending = yield from asyncio.wait(
            [
                stop_future,
                conn_future,
            ],
            return_when=asyncio.FIRST_COMPLETED
        )

        for fut in pending:
            if fut != stop_future:
                fut.cancel()

        if conn_future in done:
            endpoint = conn_future.result()
        else:
            return

        self.logger.debug("connection established!")

        while True:
            try:
                conn_future = asyncio.ensure_future(
                    endpoint.poll_start()
                )

                done, pending = yield from asyncio.wait(
                    [
                        conn_future,
                        stop_future,
                    ],
                    return_when=asyncio.FIRST_COMPLETED
                )

                for fut in pending:
                    if fut != stop_future:
                        fut.cancel()

                if conn_future in done:
                    response, data = conn_future.result()
                else:
                    return

                self._update_jobs(data["jobs"])
                self._update_trayicon_from_jobs()
                self._update_error_conditions_from_jobs()

                conn_future = asyncio.ensure_future(
                    endpoint.poll_finalise()
                )

                done, pending = yield from asyncio.wait(
                    [
                        conn_future,
                        stop_future,
                    ],
                    return_when=asyncio.FIRST_COMPLETED
                )

                for fut in pending:
                    if fut != stop_future:
                        fut.cancel()

                if conn_future in done:
                    conn_future.result()
                else:
                    return
            except ConnectionError as exc:
                self.logger.error(
                    "connection to scheduler broke: %s",
                    exc
                )
                break

    @asyncio.coroutine
    def run(self):
        self.trayicon.show()
        stop_future = asyncio.ensure_future(self.stop_event.wait())

        yield from try_start()

        while not self._quit:
            yield from self._update_loop(stop_future)

        if self._stop:
            yield from try_stop()



#pylint: skip-file
# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'data/ui/manager/ui_main.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(786, 521)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.tabWidget = QtWidgets.QTabWidget(self.centralwidget)
        self.tabWidget.setObjectName("tabWidget")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.tab)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.repositories_view = QtWidgets.QTableView(self.tab)
        self.repositories_view.setAlternatingRowColors(True)
        self.repositories_view.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.repositories_view.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.repositories_view.setShowGrid(False)
        self.repositories_view.setSortingEnabled(True)
        self.repositories_view.setWordWrap(False)
        self.repositories_view.setCornerButtonEnabled(False)
        self.repositories_view.setObjectName("repositories_view")
        self.repositories_view.horizontalHeader().setHighlightSections(False)
        self.repositories_view.horizontalHeader().setSortIndicatorShown(True)
        self.repositories_view.horizontalHeader().setStretchLastSection(True)
        self.repositories_view.verticalHeader().setVisible(False)
        self.verticalLayout_2.addWidget(self.repositories_view)
        self.tabWidget.addTab(self.tab, "")
        self.tab_2 = QtWidgets.QWidget()
        self.tab_2.setObjectName("tab_2")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.tab_2)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.jobs_view = QtWidgets.QTableView(self.tab_2)
        self.jobs_view.setAlternatingRowColors(True)
        self.jobs_view.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.jobs_view.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.jobs_view.setShowGrid(False)
        self.jobs_view.setSortingEnabled(True)
        self.jobs_view.setWordWrap(False)
        self.jobs_view.setObjectName("jobs_view")
        self.jobs_view.horizontalHeader().setHighlightSections(False)
        self.jobs_view.horizontalHeader().setSortIndicatorShown(True)
        self.jobs_view.horizontalHeader().setStretchLastSection(True)
        self.jobs_view.verticalHeader().setVisible(False)
        self.verticalLayout_3.addWidget(self.jobs_view)
        self.tabWidget.addTab(self.tab_2, "")
        self.verticalLayout.addWidget(self.tabWidget)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 786, 25))
        self.menubar.setObjectName("menubar")
        self.menuRepository = QtWidgets.QMenu(self.menubar)
        self.menuRepository.setObjectName("menuRepository")
        self.menuJob = QtWidgets.QMenu(self.menubar)
        self.menuJob.setObjectName("menuJob")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.action_repository_add = QtWidgets.QAction(MainWindow)
        self.action_repository_add.setObjectName("action_repository_add")
        self.action_job_new = QtWidgets.QAction(MainWindow)
        self.action_job_new.setObjectName("action_job_new")
        self.menuRepository.addAction(self.action_repository_add)
        self.menuJob.addAction(self.action_job_new)
        self.menubar.addAction(self.menuRepository.menuAction())
        self.menubar.addAction(self.menuJob.menuAction())

        self.retranslateUi(MainWindow)
        self.tabWidget.setCurrentIndex(1)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "CuteBorg Manager"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), _translate("MainWindow", "Repositories"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_2), _translate("MainWindow", "Jobs"))
        self.menuRepository.setTitle(_translate("MainWindow", "Repositor&y"))
        self.menuJob.setTitle(_translate("MainWindow", "&Job"))
        self.action_repository_add.setText(_translate("MainWindow", "&Add repository…"))
        self.action_repository_add.setToolTip(_translate("MainWindow", "Add an existing local or remote repository."))
        self.action_job_new.setText(_translate("MainWindow", "&New job…"))
        self.action_job_new.setToolTip(_translate("MainWindow", "Create a new backup job."))




########################################################################
# File name: utils.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import calendar

from datetime import datetime, timedelta

from .. import config


def align_to_minutes(timestamp, nminutes):
    """
    Align the given `timestamp` to `nminutes`.

    The alignment base is determined by the following rules, in order:

    * If 60 is divisible by `nminutes`, the timestamp is aligned based on the
      hour.

    * If 1440 is divisible by `nminutes`, the timestamp is aligned based on the
      day.

    * Otherwise, the timestamp is aligned based on the UNIX epoch
      (1970-01-01T01:00:00Z).

    `timestamp` must not have a timezone attached, so that timedelta operations
    work. It is assumed to be in UTC.
    """
    if timestamp.tzinfo is not None:
        raise ValueError(
            "align_to_minutes needs naive datetime"
        )

    if 60 % nminutes == 0:
        # nminutes can be aligned by hour
        timestamp = timestamp.replace(
            minute=timestamp.minute // nminutes * nminutes,
            second=0,
            microsecond=0,
        )

    elif 1440 % nminutes == 0:
        # nminutes can be aligned by day
        minute_of_day = timestamp.hour * 60 + timestamp.minute
        hour, minute = divmod(
            minute_of_day // nminutes * nminutes,
            60,
        )

        timestamp = timestamp.replace(
            hour=hour,
            minute=minute,
            second=0,
            microsecond=0,
        )

    else:
        # we need to operate on UNIX epoch
        ts = calendar.timegm(timestamp.timetuple())
        mod = nminutes*60
        ts = ts // mod * mod
        timestamp = datetime.utcfromtimestamp(ts)

    return timestamp


def align_to_hours(timestamp, nhours):
    """
    Align the given `timestamp` to `nhours`.

    If 24 is divisible by `nhours`, the timestamp is aligned based on the day.
    Otherwise, it is based on the UNIX epoch (1970-01-01T01:00:00Z).

    `timestamp` must not have a timezone attached, so that timedelta operations
    work. It is assumed to be in UTC.
    """
    if timestamp.tzinfo is not None:
        raise ValueError(
            "align_to_hours needs naive datetime"
        )

    if 24 % nhours == 0:
        # nhours can be aligned by day
        timestamp = timestamp.replace(
            hour=timestamp.hour // nhours * nhours,
            minute=0,
            second=0,
            microsecond=0,
        )
    else:
        # we need to operate on UNIX epoch
        ts = calendar.timegm(timestamp.timetuple())
        mod = nhours*3600
        ts = ts // mod * mod
        timestamp = datetime.utcfromtimestamp(ts)

    return timestamp


def align_to_days(timestamp, ndays):
    """
    Align the given `timestamp` to `ndays`.

    Day alignment is always relative to the UNIX epoch (1970-01-01T01:00:00Z).

    `timestamp` must not have a timezone attached, so that timedelta operations
    work. It is assumed to be in UTC.
    """
    if timestamp.tzinfo is not None:
        raise ValueError(
            "align_to_days needs naive datetime"
        )

    # we need to operate on UNIX epoch
    ts = calendar.timegm(timestamp.timetuple())
    mod = ndays*86400
    ts = ts // mod * mod
    timestamp = datetime.utcfromtimestamp(ts)

    return timestamp


def align_to_weeks(timestamp, nweeks):
    """
    Align the given `timestamp` to `nweeks`.

    Week alignment is always relative to the first monday of the UNIX epoch
    (1970-01-05T00:00:00Z).

    `timestamp` must not have a timezone attached, so that timedelta operations
    work. It is assumed to be in UTC.
    """
    if timestamp.tzinfo is not None:
        raise ValueError(
            "align_to_days needs naive datetime"
        )

    mod = nweeks*7*86400

    # we need to operate on UNIX epoch
    ts = calendar.timegm(timestamp.timetuple())
    # move back four days to end up at the actual epoch
    ts -= (4*86400)
    ts = ts // mod * mod
    ts += (4*86400)
    timestamp = datetime.utcfromtimestamp(ts)

    return timestamp


def align_to_months(timestamp, nmonths):
    """
    Align the given `timestamp` to `nmonths`.

    If 12 is divisible by `nmonths`, the alignment is based on the year.
    Otherwise, it is based on the UNIX epoch (1970-01-01T01:00:00Z).

    `timestamp` must not have a timezone attached, so that timedelta operations
    work. It is assumed to be in UTC.
    """
    if timestamp.tzinfo is not None:
        raise ValueError(
            "align_to_months needs naive datetime"
        )

    if 12 % nmonths == 0:
        timestamp = timestamp.replace(
            month=((timestamp.month-1) // nmonths * nmonths)+1,
            day=1,
            hour=0,
            minute=0,
            second=0,
            microsecond=0,
        )

    else:
        months_since_epoch = (timestamp.year - 1970) * 12 + timestamp.month-1
        months_since_epoch = months_since_epoch // nmonths * nmonths

        year, month = divmod(months_since_epoch, 12)

        timestamp = timestamp.replace(
            year=1970+year,
            month=month+1,
            day=1,
            hour=0,
            minute=0,
            second=0,
            microsecond=0,
        )

    return timestamp


def align_to_years(timestamp, nyears):
    """
    Align the given `timestamp` to `nyears`.

    Year alignment is always based on the UNIX epoch (1970-01-01T01:00:00Z).

    `timestamp` must not have a timezone attached, so that timedelta operations
    work. It is assumed to be in UTC.
    """
    if timestamp.tzinfo is not None:
        raise ValueError(
            "align_to_months needs naive datetime"
        )

    timestamp = timestamp.replace(
        year=(timestamp.year-1970) // nyears * nyears + 1970,
        month=1,
        day=1,
        hour=0,
        minute=0,
        second=0,
        microsecond=0,
    )

    return timestamp


def align_to_interval(timestamp, interval_step, interval_unit):
    """
    Align `timestamp` to a given interval.

    :param timestamp: Timestamp to align
    :type timestamp: :class:`datetime.datetime`
    :param interval_step: Length of the interval, in units of `interval_unit`.
    :type interval_step: :class:`int`
    :param interval_unit: Step size of the interval
    :type interval_unit: :class:`~.IntervalUnit`
    :raises ValueError: If `timestamp` is not a naive
                        :class:`~datetime.datetime`.
    :return: Aligned timestamp
    :rtype: :class:`datetime.datetime`
    """

    if timestamp.tzinfo is not None:
        raise ValueError(
            "align_to_interval needs naive datetime"
        )

    impl = {
        config.IntervalUnit.MINUTE: align_to_minutes,
        config.IntervalUnit.HOUR: align_to_hours,
        config.IntervalUnit.DAY: align_to_days,
        config.IntervalUnit.WEEK: align_to_weeks,
        config.IntervalUnit.MONTH: align_to_months,
        config.IntervalUnit.YEAR: align_to_years,
    }[interval_unit]

    return impl(timestamp, interval_step)


def step_interval(timestamp, interval_step, interval_unit):
    """
    Step a timestamp aligned to the interval by one interval.

    :param timestamp: Timestamp aligned to the interval
    :type timestamp: :class:`datetime.datetime`
    :param interval_step: Length of the interval, in units of `interval_unit`.
    :type interval_step: :class:`int`
    :param interval_unit: Unit of the interval steps.
    :type interval_unit: :class:`~.IntervalUnit`
    :raises ValueError: If `interval_unit` is invalid.
    :raises ValueError: If `timestamp` is not a naive
                        :class:`~datetime.datetime`.
    :raises ValueError: If the resulting timestamp is not valid; this can
                        happen if `timestamp` is not aligned to the interval.
    :return: Timestamp at the beginning of the next interval.
    :rtype: :class:`datetime.datetime`

    .. note::

       The `timestamp` must already be aligned. If it isn’t aligned, the
       resulting timestamp is also not aligned.

       Also, in the case of some interval units, it is possible that the
       resulting timestamp is invalid, thus raising a :class:`ValueError`.
    """

    if timestamp.tzinfo is not None:
        raise ValueError(
            "step_interval needs naive datetime"
        )

    if interval_unit == config.IntervalUnit.MINUTE:
        return timestamp + timedelta(seconds=60*interval_step)

    elif interval_unit == config.IntervalUnit.HOUR:
        return timestamp + timedelta(seconds=3600*interval_step)

    elif interval_unit == config.IntervalUnit.DAY:
        return timestamp + timedelta(days=interval_step)

    elif interval_unit == config.IntervalUnit.WEEK:
        return timestamp + timedelta(days=7*interval_step)

    elif interval_unit == config.IntervalUnit.MONTH:
        months_total = timestamp.year * 12 + timestamp.month
        months_total += interval_step
        year, month = divmod(months_total, 12)
        return timestamp.replace(
            year=year,
            month=month,
        )

    elif interval_unit == config.IntervalUnit.YEAR:
        return timestamp.replace(
            year=timestamp.year+interval_step,
        )

    raise ValueError(
        "invalid interval_unit: {!r}".format(interval_unit)
    )
